@@ -1,14 +1,15 @@
 /*
  * pg_reorg.c: bin/pg_reorg.c
  *
- * Copyright (c) 2008-2010, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
+ * Portions Copyright (c) 2008-2011, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
+ * Portions Copyright (c) 2011, Itagaki Takahiro
  */
 
 /**
  * @brief Client Modules
  */
 
-const char *PROGRAM_VERSION	= "1.1.5";
+const char *PROGRAM_VERSION	= "1.1.8";
 const char *PROGRAM_URL		= "http://reorg.projects.postgresql.org/";
 const char *PROGRAM_EMAIL	= "reorg-general@lists.pgfoundry.org";
 
@@ -19,21 +20,29 @@ const char *PROGRAM_EMAIL	= "reorg-general@lists.pgfoundry.org";
 #include <unistd.h>
 #include <time.h>
 
+/*
+ * APPLY_COUNT: Number of applied logs per transaction. Larger values
+ * could be faster, but will be long transactions in the REDO phase.
+ */
 #define APPLY_COUNT		1000
 
+/* The '1/1, -1/0' lock skipped is from the bgwriter on newly promoted
+ * servers. See GH ticket #1.
+ */
 #define SQL_XID_SNAPSHOT_80300 \
 	"SELECT reorg.array_accum(virtualtransaction) FROM pg_locks"\
-	" WHERE locktype = 'virtualxid' AND pid <> pg_backend_pid()"
+	" WHERE locktype = 'virtualxid' AND pid <> pg_backend_pid()"\
+	" AND (virtualxid, virtualtransaction) <> ('1/1', '-1/0')"
 #define SQL_XID_SNAPSHOT_80200 \
 	"SELECT reorg.array_accum(transactionid) FROM pg_locks"\
 	" WHERE locktype = 'transactionid' AND pid <> pg_backend_pid()"
 
 #define SQL_XID_ALIVE_80300 \
-	"SELECT 1 FROM pg_locks WHERE locktype = 'virtualxid'"\
-	" AND pid <> pg_backend_pid() AND virtualtransaction = ANY($1) LIMIT 1"
+	"SELECT pid FROM pg_locks WHERE locktype = 'virtualxid'"\
+	" AND pid <> pg_backend_pid() AND virtualtransaction = ANY($1)"
 #define SQL_XID_ALIVE_80200 \
-	"SELECT 1 FROM pg_locks WHERE locktype = 'transactionid'"\
-	" AND pid <> pg_backend_pid() AND transactionid = ANY($1) LIMIT 1"
+	"SELECT pid FROM pg_locks WHERE locktype = 'transactionid'"\
+	" AND pid <> pg_backend_pid() AND transactionid = ANY($1)"
 
 #define SQL_XID_SNAPSHOT \
 	(PQserverVersion(connection) >= 80300 \
@@ -60,6 +69,7 @@ typedef struct reorg_table
 	const char	   *create_log;		/* CREATE TABLE log */
 	const char	   *create_trigger;	/* CREATE TRIGGER z_reorg_trigger */
 	const char	   *create_table;	/* CREATE TABLE table AS SELECT */
+	const char	   *drop_columns;	/* ALTER TABLE DROP COLUMNs */
 	const char	   *delete_log;		/* DELETE FROM log */
 	const char	   *lock_table;		/* LOCK TABLE table */
 	const char	   *sql_peek;		/* SQL used in flush */
@@ -293,13 +303,14 @@ reorg_one_database(const char *orderby, const char *table)
 		if (table.pkid == 0)
 			ereport(ERROR,
 				(errcode(E_PG_COMMAND),
-				 errmsg("relation \"%s\" has no primary key", table.target_name)));
+				 errmsg("relation \"%s\" must have a primary key or not-null unique keys", table.target_name)));
 
 		table.create_pktype = getstr(res, i, c++);
 		table.create_log = getstr(res, i, c++);
 		table.create_trigger = getstr(res, i, c++);
 
 		create_table = getstr(res, i, c++);
+		table.drop_columns = getstr(res, i, c++);
 		table.delete_log = getstr(res, i, c++);
 		table.lock_table = getstr(res, i, c++);
 		ckey = getstr(res, i, c++);
@@ -376,6 +387,7 @@ reorg_one_table(const reorg_table *table, const char *orderby)
 	const char	   *params[1];
 	int				num;
 	int				i;
+	int				num_waiting = 0;
 	char		   *vxid;
 	char			buffer[12];
 	StringInfoData	sql;
@@ -393,6 +405,7 @@ reorg_one_table(const reorg_table *table, const char *orderby)
 	elog(DEBUG2, "create_log     : %s", table->create_log);
 	elog(DEBUG2, "create_trigger : %s", table->create_trigger);
 	elog(DEBUG2, "create_table   : %s", table->create_table);
+	elog(DEBUG2, "drop_columns   : %s", table->drop_columns ? table->drop_columns : "(skipped)");
 	elog(DEBUG2, "delete_log     : %s", table->delete_log);
 	elog(DEBUG2, "lock_table     : %s", table->lock_table);
 	elog(DEBUG2, "sql_peek       : %s", table->sql_peek);
@@ -419,6 +432,7 @@ reorg_one_table(const reorg_table *table, const char *orderby)
 			(errcode(E_PG_COMMAND),
 			 errmsg("trigger %s conflicted for %s",
 					PQgetvalue(res, 0, 0), table->target_name)));
+	PQclear(res);
 
 	command(table->create_pktype, 0, NULL);
 	command(table->create_log, 0, NULL);
@@ -430,7 +444,7 @@ reorg_one_table(const reorg_table *table, const char *orderby)
 	/*
 	 * Register the table to be dropped on error. We use pktype as
 	 * an advisory lock. The registration should be done after
-	 * the first command is succeeded.
+	 * the first command succeeds.
 	 */
 	pgut_atexit_push(&reorg_cleanup, (void *) table);
 
@@ -450,6 +464,8 @@ reorg_one_table(const reorg_table *table, const char *orderby)
 	command(table->delete_log, 0, NULL);
 	command(table->create_table, 0, NULL);
 	printfStringInfo(&sql, "SELECT reorg.disable_autovacuum('reorg.table_%u')", table->target_oid);
+	if (table->drop_columns)
+		command(table->drop_columns, 0, NULL);
 	command(sql.data, 0, NULL);
 	command("COMMIT", 0, NULL);
 
@@ -460,7 +476,9 @@ reorg_one_table(const reorg_table *table, const char *orderby)
 
 	params[0] = utoa(table->target_oid, buffer);
 	res = execute("SELECT indexrelid,"
-		" reorg.reorg_indexdef(indexrelid, indrelid)"
+		" reorg.reorg_indexdef(indexrelid, indrelid),"
+		" indisvalid,"
+		" pg_get_indexdef(indexrelid)"
 		" FROM pg_index WHERE indrelid = $1", 1, params);
 
 	num = PQntuples(res);
@@ -468,9 +486,18 @@ reorg_one_table(const reorg_table *table, const char *orderby)
 	{
 		reorg_index	index;
 		int			c = 0;
+		const char *isvalid;
+		const char *indexdef;
 
 		index.target_oid = getoid(res, i, c++);
 		index.create_index = getstr(res, i, c++);
+		isvalid = getstr(res, i, c++);
+		indexdef = getstr(res, i, c++);
+
+		if (isvalid && isvalid[0] == 'f') {
+			elog(WARNING, "skipping invalid index: %s", indexdef);
+			continue;
+		}
 
 		elog(DEBUG2, "[%d]", i);
 		elog(DEBUG2, "target_oid   : %u", index.target_oid);
@@ -485,8 +512,8 @@ reorg_one_table(const reorg_table *table, const char *orderby)
 	PQclear(res);
 
 	/*
-	 * 4. Apply log to temp table until no tuples left in the log
-	 * and all of old transactions are finished.
+	 * 4. Apply log to temp table until no tuples are left in the log
+	 * and all of the old transactions are finished.
 	 */
 	for (;;)
 	{
@@ -498,16 +525,32 @@ reorg_one_table(const reorg_table *table, const char *orderby)
 		params[0] = vxid;
 		res = execute(SQL_XID_ALIVE, 1, params);
 		num = PQntuples(res);
-		PQclear(res);
 
 		if (num > 0)
 		{
-			sleep(1);
-			continue;	/* wait for old transactions */
-		}
+			/* Wait for old transactions.
+			 * Only display the message below when the number of
+			 * transactions we are waiting on changes (presumably,
+			 * num_waiting should only go down), so as not to
+			 * be too noisy.
+			 */
+			if (num != num_waiting)
+			{
+				elog(NOTICE, "Waiting for %d transactions to finish. First PID: %s", num, PQgetvalue(res, 0, 0));
+				num_waiting = num;
+			}
 
-		/* ok, go next step. */
-		break;
+			PQclear(res);
+			sleep(1);
+			continue;
+		}
+		else
+		{
+			/* All old transactions are finished;
+			 * go to next step. */
+			PQclear(res);
+			break;
+		}
 	}
 
 	/*
@@ -635,7 +678,7 @@ reorg_cleanup(bool fatal, void *userdata)
 
 	if (fatal)
 	{
-		fprintf(stderr, "!!!FATAL ERROR!!! Please refer to a manual.\n\n");
+		fprintf(stderr, "!!!FATAL ERROR!!! Please refer to the manual.\n\n");
 	}
 	else
 	{
@@ -668,9 +711,9 @@ pgut_help(bool details)
 
 	printf("Options:\n");
 	printf("  -a, --all                 reorg all databases\n");
-	printf("  -t, --table=TABLE         reorg specific table only\n");
 	printf("  -n, --no-order            do vacuum full instead of cluster\n");
 	printf("  -o, --order-by=columns    order by columns instead of cluster keys\n");
-	printf("  -T, --wait-timeout=secs   timeout to cancel other backends on conflict.\n");
+	printf("  -t, --table=TABLE         reorg specific table only\n");
+	printf("  -T, --wait-timeout=secs   timeout to cancel other backends on conflict\n");
 	printf("  -Z, --no-analyze          don't analyze at end\n");
 }
