@@ -55,6 +55,11 @@ CREATE TABLE tbl_with_dropped_toast (
 );
 ALTER TABLE tbl_with_dropped_toast CLUSTER ON tbl_with_dropped_toast_pkey;
 
+CREATE TABLE tbl_badindex (
+	id integer PRIMARY KEY,
+	n integer
+);
+
 --
 -- insert data
 --
@@ -80,15 +85,27 @@ ALTER TABLE tbl_with_dropped_column DROP COLUMN d1;
 ALTER TABLE tbl_with_dropped_column DROP COLUMN d2;
 ALTER TABLE tbl_with_dropped_column DROP COLUMN d3;
 ALTER TABLE tbl_with_dropped_column ADD COLUMN c3 text;
+CREATE VIEW view_for_dropped_column AS
+	SELECT * FROM tbl_with_dropped_column;
 
 INSERT INTO tbl_with_dropped_toast VALUES(1, 10, 'abc');
 INSERT INTO tbl_with_dropped_toast VALUES(2, 20, sqrt(2::numeric(1000,999))::text || sqrt(3::numeric(1000,999))::text);
 ALTER TABLE tbl_with_dropped_toast DROP COLUMN t;
+
+INSERT INTO tbl_badindex VALUES(1, 10);
+INSERT INTO tbl_badindex VALUES(2, 10);
+
+-- This will fail. Silence the message as it's different across PG versions.
+SET client_min_messages = fatal;
+CREATE UNIQUE INDEX CONCURRENTLY idx_badindex_n ON tbl_badindex (n);
+SET client_min_messages = warning;
+
 --
 -- before
 --
 
 SELECT * FROM tbl_with_dropped_column;
+SELECT * FROM view_for_dropped_column;
 SELECT * FROM tbl_with_dropped_toast;
 
 --
@@ -114,8 +131,19 @@ SELECT col1, to_char("time", 'YYYY-MM-DD HH24:MI:SS'), ","")" FROM tbl_cluster;
 SELECT * FROM tbl_only_ckey ORDER BY 1;
 SELECT * FROM tbl_only_pkey ORDER BY 1;
 SELECT * FROM tbl_gistkey ORDER BY 1;
+
+SET enable_seqscan = on;
+SET enable_indexscan = off;
 SELECT * FROM tbl_with_dropped_column;
+SELECT * FROM view_for_dropped_column;
 SELECT * FROM tbl_with_dropped_toast;
+SET enable_seqscan = off;
+SET enable_indexscan = on;
+SELECT * FROM tbl_with_dropped_column;
+SELECT * FROM view_for_dropped_column;
+SELECT * FROM tbl_with_dropped_toast;
+RESET enable_seqscan;
+RESET enable_indexscan;
 
 --
 -- check broken links or orphan toast relations
@@ -130,3 +158,23 @@ SELECT oid, relname
  WHERE relkind = 'r'
    AND reltoastrelid <> 0
    AND reltoastrelid NOT IN (SELECT oid FROM pg_class WHERE relkind = 't');
+
+--
+-- NOT NULL UNIQUE
+--
+CREATE TABLE tbl_nn    (col1 int NOT NULL, col2 int NOT NULL);
+CREATE TABLE tbl_uk    (col1 int NOT NULL, col2 int         , UNIQUE(col1, col2));
+CREATE TABLE tbl_nn_uk (col1 int NOT NULL, col2 int NOT NULL, UNIQUE(col1, col2));
+CREATE TABLE tbl_pk_uk (col1 int NOT NULL, col2 int NOT NULL, PRIMARY KEY(col1, col2), UNIQUE(col2, col1));
+CREATE TABLE tbl_nn_puk (col1 int NOT NULL, col2 int NOT NULL);
+CREATE UNIQUE INDEX tbl_nn_puk_pcol1_idx ON tbl_nn_puk(col1) WHERE col1 < 10;
+\! pg_reorg --dbname=contrib_regression --no-order --table=tbl_nn
+-- => ERROR
+\! pg_reorg --dbname=contrib_regression --no-order --table=tbl_uk
+-- => ERROR
+\! pg_reorg --dbname=contrib_regression --no-order --table=tbl_nn_uk
+-- => OK
+\! pg_reorg --dbname=contrib_regression --no-order --table=tbl_pk_uk
+-- => OK
+\! pg_reorg --dbname=contrib_regression --no-order --table=tbl_nn_puk
+-- => ERROR
