@@ -1,7 +1,8 @@
 /*
  * pg_reorg: lib/reorg.c
  *
- * Copyright (c) 2008-2010, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
+ * Portions Copyright (c) 2008-2011, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
+ * Portions Copyright (c) 2011, Itagaki Takahiro
  */
 
 #include "postgres.h"
@@ -22,12 +23,18 @@
 #include "miscadmin.h"
 #include "utils/builtins.h"
 #include "utils/lsyscache.h"
+#include "utils/rel.h"
 #include "utils/relcache.h"
 #include "utils/syscache.h"
 
 #include "pgut/pgut-spi.h"
 #include "pgut/pgut-be.h"
 
+/* htup.h was reorganized for 9.3, so now we need this header */
+#if PG_VERSION_NUM >= 90300
+#include "access/htup_details.h"
+#endif
+
 PG_MODULE_MAGIC;
 
 extern Datum PGUT_EXPORT reorg_version(PG_FUNCTION_ARGS);
@@ -68,14 +75,29 @@ must_be_superuser(const char *func)
 		elog(ERROR, "must be superuser to use %s function", func);
 }
 
+
+/* Include an implementation of RenameRelationInternal for old
+ * versions which don't have one.
+ */
 #if PG_VERSION_NUM < 80400
 static void RenameRelationInternal(Oid myrelid, const char *newrelname, Oid namespaceId);
 #endif
 
+
+/* The API of RenameRelationInternal() was changed in 9.2.
+ * Use the RENAME_REL macro for compatibility across versions.
+ */
+#if PG_VERSION_NUM < 90200
+#define RENAME_REL(relid, newrelname) RenameRelationInternal(relid, newrelname, PG_TOAST_NAMESPACE);
+#else
+#define RENAME_REL(relid, newrelname) RenameRelationInternal(relid, newrelname);
+#endif
+
+
 Datum
 reorg_version(PG_FUNCTION_ARGS)
 {
-	return CStringGetTextDatum("pg_reorg 1.1.5");
+	return CStringGetTextDatum("pg_reorg 1.1.8");
 }
 
 /**
@@ -263,7 +285,7 @@ reorg_apply(PG_FUNCTION_ARGS)
 }
 
 /*
- * Parsed CREATE INDEX statement. You can rebuild sql using 
+ * Parsed CREATE INDEX statement. You can rebuild sql using
  * sprintf(buf, "%s %s ON %s USING %s (%s)%s",
  *		create, index, table type, columns, options)
  */
@@ -504,7 +526,7 @@ reorg_get_index_keys(PG_FUNCTION_ARGS)
 			HeapTuple		tp;
 			Form_pg_opclass	opclassTup;
 #endif
-			
+
 			opclass = OpclassnameGetOpcid(BTREE_AM_OID, opcname);
 
 #if PG_VERSION_NUM >= 80300
@@ -584,109 +606,6 @@ getoid(HeapTuple tuple, TupleDesc desc, int column)
 	return isnull ? InvalidOid : DatumGetObjectId(datum);
 }
 
-static int16
-getint16(HeapTuple tuple, TupleDesc desc, int column)
-{
-	bool	isnull;
-	Datum	datum = SPI_getbinval(tuple, desc, column, &isnull);
-	return isnull ? 0 : DatumGetInt16(datum);
-}
-
-static void
-remove_dropped_columns_and_adjust_attnum(Oid oid, int16 natts1, int16 natts2)
-{
-	/* delete dropped columns */
-	execute_with_format(SPI_OK_DELETE,
-		"DELETE FROM pg_catalog.pg_attribute"
-		" WHERE attrelid = %u AND attisdropped",
-		oid);
-	if (SPI_processed != natts1 - natts2)
-		elog(ERROR, "cannot remove %d dropped columns (%u columns removed)",
-			natts2 - natts1, SPI_processed);
-
-	/* renumber attnum */
-#if PG_VERSION_NUM >= 90000
-	execute_with_format(SPI_OK_UPDATE,
-		"UPDATE pg_catalog.pg_attribute m"
-		"   SET attnum = (SELECT count(*) FROM pg_attribute a"
-		" WHERE m.attrelid = a.attrelid"
-		"   AND m.attnum >= a.attnum"
-		"   AND a.attnum > 0 AND NOT a.attisdropped)"
-		" WHERE attrelid = %u AND attnum > 0 AND NOT attisdropped",
-		oid);
-	if (SPI_processed != natts2)
-		elog(ERROR, "cannot update %d columns (%u columns updated)",
-			natts2, SPI_processed);
-#elif PG_VERSION_NUM >= 80300
-	execute_with_format(SPI_OK_UPDATE,
-		"UPDATE pg_catalog.pg_attribute"
-		"   SET attnum = (SELECT count(*) FROM pg_attribute a"
-		" WHERE pg_catalog.pg_attribute.attrelid = a.attrelid"
-		"   AND pg_catalog.pg_attribute.attnum >= a.attnum"
-		"   AND a.attnum > 0 AND NOT a.attisdropped)"
-		" WHERE attrelid = %u AND attnum > 0 AND NOT attisdropped",
-		oid);
-	if (SPI_processed != natts2)
-		elog(ERROR, "cannot update %d columns (%u columns updated)",
-			natts2, SPI_processed);
-#else
-	/*
-	 * Use count(*) in subquery because 8.2 doesn't support aggregates
-	 * in UPDATE SET.
-	 */
-	do
-	{
-		uint32			i;
-		uint32			ntuples;
-		SPITupleTable  *tuptable;
-		TupleDesc		desc;
-
-		execute_with_format(SPI_OK_SELECT,
-			"SELECT attnum FROM pg_catalog.pg_attribute"
-			" WHERE attrelid = %u AND attnum > 0 AND NOT attisdropped"
-			" ORDER BY attnum",
-			oid);
-		if (SPI_processed != natts2)
-			elog(ERROR, "number of columns should be %d (%d returned)",
-				natts2, SPI_processed);
-
-		ntuples = SPI_processed;
-		tuptable = SPI_tuptable;
-		desc = tuptable->tupdesc;
-
-		for (i = 0; i < ntuples; i++)
-		{
-			int		attnum;
-			int		count;
-
-			attnum = getint16(tuptable->vals[i], desc, 1);
-
-			execute_with_format(SPI_OK_SELECT,
-				"SELECT count(*)::smallint FROM pg_catalog.pg_attribute"
-				" WHERE attrelid = %u AND attnum > 0 AND attnum <= %d",
-				oid, attnum);
-			if (SPI_processed != 1)
-				elog(ERROR, "cannot adjust column %d", attnum);
-
-			count = getint16(SPI_tuptable->vals[0], SPI_tuptable->tupdesc, 1);
-
-			execute_with_format(SPI_OK_UPDATE,
-				"UPDATE pg_catalog.pg_attribute"
-				"   SET attnum = %d"
-				" WHERE attrelid = %u AND attnum = %d",
-				count, oid, attnum);
-			if (SPI_processed != 1)
-				elog(ERROR, "cannot update column %d", attnum);
-		}
-	} while(0);
-#endif
-
-	/* adjust attribute number of the table */
-	execute_with_format(SPI_OK_UPDATE,
-		"UPDATE pg_catalog.pg_class SET relnatts = %d WHERE oid = %u",
-		natts2, oid);
-}
-
 /**
  * @fn      Datum reorg_swap(PG_FUNCTION_ARGS)
  * @brief   Swapping relfilenode of tables and relation ids of toast tables
@@ -721,8 +640,6 @@ reorg_swap(PG_FUNCTION_ARGS)
 	Oid				reltoastidxid2;
 	Oid				owner1;
 	Oid				owner2;
-	int16			natts1;
-	int16			natts2;
 
 	/* authority check */
 	must_be_superuser("reorg_swap");
@@ -734,8 +651,7 @@ reorg_swap(PG_FUNCTION_ARGS)
 	values[0] = ObjectIdGetDatum(oid);
 	execute_with_args(SPI_OK_SELECT,
 		"SELECT X.reltoastrelid, TX.reltoastidxid, X.relowner,"
-		"       Y.oid, Y.reltoastrelid, TY.reltoastidxid, Y.relowner,"
-		"       X.relnatts, Y.relnatts"
+		"       Y.oid, Y.reltoastrelid, TY.reltoastidxid, Y.relowner"
 		"  FROM pg_catalog.pg_class X LEFT JOIN pg_catalog.pg_class TX"
 		"         ON X.reltoastrelid = TX.oid,"
 		"       pg_catalog.pg_class Y LEFT JOIN pg_catalog.pg_class TY"
@@ -760,13 +676,11 @@ reorg_swap(PG_FUNCTION_ARGS)
 	reltoastrelid2 = getoid(tuple, desc, 5);
 	reltoastidxid2 = getoid(tuple, desc, 6);
 	owner2 = getoid(tuple, desc, 7);
-	natts1 = getint16(tuple, desc, 8);
-	natts2 = getint16(tuple, desc, 9);;
 
 	/* change owner of new relation to original owner */
 	if (owner1 != owner2)
 	{
-		ATExecChangeOwner(oid2, owner1, true);
+		ATExecChangeOwner(oid2, owner1, true, AccessExclusiveLock);
 		CommandCounterIncrement();
 	}
 
@@ -783,6 +697,7 @@ reorg_swap(PG_FUNCTION_ARGS)
 		"       pg_catalog.pg_class Y"
 		" WHERE I.indrelid = $1"
 		"   AND I.indexrelid = X.oid"
+		"   AND I.indisvalid"
 		"   AND Y.oid = ('reorg.index_' || X.oid)::regclass",
 		1, argtypes, values, nulls);
 
@@ -799,27 +714,6 @@ reorg_swap(PG_FUNCTION_ARGS)
 		idx2 = getoid(tuple, desc, 2);
 		swap_heap_or_index_files(idx1, idx2);
 
-		/* adjust key attnum if the target table has dropped columns */
-		if (natts1 != natts2)
-		{
-#if PG_VERSION_NUM >= 90000
-			execute_with_format(SPI_OK_UPDATE,
-				"UPDATE pg_catalog.pg_index m SET indkey = n.indkey"
-				"  FROM pg_catalog.pg_index n"
-				" WHERE m.indexrelid = %u"
-				"   AND n.indexrelid = 'reorg.index_%u'::regclass",
-				idx1, idx1);
-#else
-			execute_with_format(SPI_OK_UPDATE,
-				"UPDATE pg_catalog.pg_index SET indkey = n.indkey"
-				"  FROM pg_catalog.pg_index n"
-				" WHERE pg_catalog.pg_index.indexrelid = %u"
-				"   AND n.indexrelid = 'reorg.index_%u'::regclass",
-				idx1, idx1);
-#endif
-			if (SPI_processed != 1)
-				elog(ERROR, "failed to update pg_index.indkey (%u rows updated)", SPI_processed);
-		}
 		CommandCounterIncrement();
 	}
 
@@ -844,9 +738,9 @@ reorg_swap(PG_FUNCTION_ARGS)
 
 		/* rename X to Y */
 		snprintf(name, NAMEDATALEN, "pg_toast_%u", oid2);
-		RenameRelationInternal(reltoastrelid1, name, PG_TOAST_NAMESPACE);
+		RENAME_REL(reltoastrelid1, name);
 		snprintf(name, NAMEDATALEN, "pg_toast_%u_index", oid2);
-		RenameRelationInternal(reltoastidxid1, name, PG_TOAST_NAMESPACE);
+		RENAME_REL(reltoastidxid1, name);
 		CommandCounterIncrement();
 	}
 	else if (reltoastrelid1 != InvalidOid)
@@ -856,30 +750,26 @@ reorg_swap(PG_FUNCTION_ARGS)
 
 		/* rename X to TEMP */
 		snprintf(name, NAMEDATALEN, "pg_toast_pid%d", pid);
-		RenameRelationInternal(reltoastrelid1, name, PG_TOAST_NAMESPACE);
+		RENAME_REL(reltoastrelid1, name);
 		snprintf(name, NAMEDATALEN, "pg_toast_pid%d_index", pid);
-		RenameRelationInternal(reltoastidxid1, name, PG_TOAST_NAMESPACE);
+		RENAME_REL(reltoastidxid1, name);
 		CommandCounterIncrement();
 
 		/* rename Y to X */
 		snprintf(name, NAMEDATALEN, "pg_toast_%u", oid);
-		RenameRelationInternal(reltoastrelid2, name, PG_TOAST_NAMESPACE);
+		RENAME_REL(reltoastrelid2, name);
 		snprintf(name, NAMEDATALEN, "pg_toast_%u_index", oid);
-		RenameRelationInternal(reltoastidxid2, name, PG_TOAST_NAMESPACE);
+		RENAME_REL(reltoastidxid2, name);
 		CommandCounterIncrement();
 
 		/* rename TEMP to Y */
 		snprintf(name, NAMEDATALEN, "pg_toast_%u", oid2);
-		RenameRelationInternal(reltoastrelid1, name, PG_TOAST_NAMESPACE);
+		RENAME_REL(reltoastrelid1, name);
 		snprintf(name, NAMEDATALEN, "pg_toast_%u_index", oid2);
-		RenameRelationInternal(reltoastidxid1, name, PG_TOAST_NAMESPACE);
+		RENAME_REL(reltoastidxid1, name);
 		CommandCounterIncrement();
 	}
 
-	/* adjust attribute numbers if the target table has dropped columns */
-	if (natts1 != natts2)
-		remove_dropped_columns_and_adjust_attnum(oid, natts1, natts2);
-
 	/* drop reorg trigger */
 	execute_with_format(
 		SPI_OK_UTILITY,
@@ -907,6 +797,12 @@ reorg_drop(PG_FUNCTION_ARGS)
 	const char *relname = get_quoted_relname(oid);
 	const char *nspname = get_quoted_nspname(oid);
 
+	if (!(relname && nspname))
+	{
+		elog(ERROR, "table name not found for OID %u", oid);
+		PG_RETURN_VOID();
+	}
+
 	/* authority check */
 	must_be_superuser("reorg_drop");
 
@@ -1005,13 +901,15 @@ reorg_prepare(const char *src, int nargs, Oid *argtypes)
 static const char *
 get_quoted_relname(Oid oid)
 {
-	return quote_identifier(get_rel_name(oid));
+	const char *relname = get_rel_name(oid);
+	return (relname ? quote_identifier(relname) : NULL);
 }
 
 static const char *
 get_quoted_nspname(Oid oid)
 {
-	return quote_identifier(get_namespace_name(get_rel_namespace(oid)));
+	const char *nspname = get_namespace_name(get_rel_namespace(oid));
+	return (nspname ? quote_identifier(nspname) : NULL);
 }
 
 /*
@@ -1075,7 +973,11 @@ swap_heap_or_index_files(Oid r1, Oid r2)
 
 	/* swap size statistics too, since new rel has freshly-updated stats */
 	{
+#if PG_VERSION_NUM >= 90300
+		int32		swap_pages;
+#else
 		int4		swap_pages;
+#endif
 		float4		swap_tuples;
 
 		swap_pages = relform1->relpages;
@@ -1121,7 +1023,8 @@ swap_heap_or_index_files(Oid r1, Oid r2)
 		if (relform1->reltoastrelid)
 		{
 			count = deleteDependencyRecordsFor(RelationRelationId,
-											   relform1->reltoastrelid);
+											   relform1->reltoastrelid,
+											   false);
 			if (count != 1)
 				elog(ERROR, "expected one dependency record for TOAST table, found %ld",
 					 count);
@@ -1129,7 +1032,8 @@ swap_heap_or_index_files(Oid r1, Oid r2)
 		if (relform2->reltoastrelid)
 		{
 			count = deleteDependencyRecordsFor(RelationRelationId,
-											   relform2->reltoastrelid);
+											   relform2->reltoastrelid,
+											   false);
 			if (count != 1)
 				elog(ERROR, "expected one dependency record for TOAST table, found %ld",
 					 count);
@@ -1181,6 +1085,7 @@ swap_heap_or_index_files(Oid r1, Oid r2)
 
 #if PG_VERSION_NUM < 80400
 
+/* XXX: You might need to add PGDLLIMPORT into your miscadmin.h. */
 extern PGDLLIMPORT bool allowSystemTableMods;
 
 static void
